#!/usr/bin/python
import os, shutil, stat, time, subprocess, sys
from collections import OrderedDict

# some utility functions
def safeCall(*args):
	res = ""
	for arg in args:
		assert arg.find("'") < 0 # ' is not supported
		if len(res): res += " "
		res += "'"+arg+"'"
	return res

# abstract representation of rules file
class RulesFile:
	def __init__(self):
		self.env = {}
		self.dh = []
		self.rules = OrderedDict()
	
	def env2str(self):
		res = ""
		for name in self.env: # we rely on the name being sane (i.e., no special characters)
			val = self.env[name]
			assert val.find("'") < 0 # ' is not supported
			if len(res): res += " "
			res += name+"='"+val+"'"
		return res

	def write(self, f):
		print >>f, "#!/usr/bin/make -f"
		print >>f, ""
		print >>f, ".PHONY: build" # there may be a directory called "build"
		print >>f, ""
		print >>f, "build %:" # need to mention "build" here again explicitly
		print >>f, '\t'+self.env2str()+' dh $@ '+safeCall(*self.dh)
		for rule in self.rules:
			print >>f, ""
			print >>f, "override_dh_"+rule+":"
			for line in self.rules[rule]:
				print >>f, "\t"+line

# build-system specific part of rules file
def cmakeRules(config):
	buildDir = config.get('buildDir', 'build')
	srcDir = os.getcwd()
	r = RulesFile()
	r.dh += ["--buildsystem=cmake", "--builddirectory="+buildDir] # dh parameters
	r.rules['auto_configure'] = [
		safeCall("mkdir", "-p", buildDir),
		safeCall("cd", buildDir) + " && " +
		  safeCall("cmake", srcDir, "-DCMAKE_INSTALL_PREFIX=/usr", *config.get('cmakeParameters', []))
	]
	r.rules['auto_clean'] = [safeCall('rm', '-f', os.path.join(buildDir, 'CMakeCache.txt'))] # clean old cmake cache
	return r

def automakeRules(config):
	# "build" is what we are building *on*, and "host" is what we are building *for* (and GNU is weird...)
	# also, we don't really support cross-building... ;-) (to do, we'd have to write shell code that checks whether BUILD_GNU_TYPE
	#  and HOST_GNU_TYPE are equal, and if they are not, add a --host parameter)
	r = RulesFile()
	r.dh += ["--buildsystem=autoconf"]
	r.rules['auto_configure'] = [
		'BUILD_TYPE=$$(dpkg-architecture -qDEB_BUILD_GNU_TYPE) && ' + # doing the expansion beforehand ensures that we cancel if it fails
		'MULTIARCH=$$(dpkg-architecture -qDEB_HOST_MULTIARCH) && '+
			'./configure --build=$$BUILD_TYPE '+ # do the escape manually here so we can use the variables (there's no user-controlled string in here anyway)
			'--prefix=/usr --includedir=/usr/include --mandir=/usr/share/man --infodir=/usr/share/info '+
			'--libdir=/usr/lib/$$MULTIARCH --libexecdir=/usr/lib/$$MULTIARCH '+
			'--sysconfdir=/etc --localstatedir=/var '+
			safeCall(*config.get('automakeParameters', []))
	]
	r.rules['auto_clean'] = ['rm -f config.status'] # do not re-use old configuration (no need for escaping here, obviously)
	return r

# utility functions
def commandInBuildEnv(config, command):
	schroot = config.get('schroot')
	if schroot is not None: command = ['schroot', '-c', schroot, '--'] + command
	return command

def getArchitecture(config):
	cmd = commandInBuildEnv(config, ['dpkg-architecture', '-qDEB_HOST_ARCH'])
	p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
	res = p.communicate()[0] # get only stdout
	if p.returncode != 0: raise Exception("Querying dpkg for the architecture failed")
	return res[0:len(res)-1] # chop of the \n at the end

def writeDependency(f, name, list):
	if len(list):
		print >>f, name+": "+', '.join(list)

# actual work functions
def createDebianFiles(config):
	sourceName = config['sourceName']
	binaryName = config.get('binaryName', sourceName+'-local')
	name = config.get('name', os.getenv('USER')) # os.getlogin() fails in minimal chroots
	email = config.get('email', os.getenv('USER')+'@'+os.uname()[1]) # user@hostname
	debDir = os.path.expanduser(config['debDir'])
	buildSystem = config['buildSystem']
	version = config['version']
	dbgPackage = config.get('dbgPackage', False)
	parallelJobs = int(config.get('parallelJobs', 2))
	packageArchitecture = config.get('architecture', 'any')
	withPython2 = config.get('withPython2', False)
	# we return the list of files generated, so we need to know the architecture
	arch = getArchitecture(config)
	files = []
	# create folders
	if os.path.exists('debian'): raise Exception('debian folder already exists?')
	os.mkdir('debian')
	os.mkdir('debian/source')
	if not os.path.exists(debDir): os.makedirs(debDir)
	# source format file
	with open('debian/source/format', 'w') as f:
		print >>f, "3.0 (native)"
	# compat file
	with open('debian/compat', 'w') as f:
		print >>f, "9"
	# copyright file
	with open('debian/copyright', 'w') as f:
		print >>f, "Auto-generated by auto-debuild, not suited for distribution"
	# changelog file
	with open('debian/changelog', 'w') as f:
		print >>f, sourceName,"("+version+")","UNRELEASED; urgency=low"
		print >>f, ""
		print >>f, "  * Auto-generated by auto-debuild"
		print >>f, ""
		print >>f, " --",name,"<"+email+">  "+time.strftime('%a, %d %b %Y %H:%M:%S %z')
	# control file
	with open('debian/control', 'w') as f:
		print >>f, "Source:",sourceName
		print >>f, "Section:",config.get('section', 'misc')
		print >>f, "Priority: extra"
		print >>f, "Maintainer: %s <%s>" % (name, email)
		writeDependency(f, 'Build-Depends', ["debhelper (>= 9)"] + config.get('buildDepends', []))
		print >>f, "Standards-Version: 3.9.3"
		print >>f, ""
		print >>f, "Package:",binaryName
		print >>f, "Architecture:",packageArchitecture
		writeDependency(f, "Pre-Depends", ["${misc:Pre-Depends}"] + config.get('binaryPreDepends', []))
		writeDependency(f, "Depends", ["${shlibs:Depends}", "${misc:Depends}"] + config.get('binaryDepends', []))
		writeDependency(f, "Recommends", config.get('binaryRecommends', []))
		writeDependency(f, "Provides", config.get('binaryProvides', [sourceName]))
		writeDependency(f, "Conflicts", config.get('binaryConflicts', []))
		print >>f, "Description:",sourceName,"(auto-debuild)"
		print >>f, " Package auto-generated by auto-debuild."
		files.append(os.path.join(debDir, "%s_%s_%s.deb" % (binaryName, version, arch)))
		if dbgPackage:
			print >>f, ""
			print >>f, "Package:",binaryName+"-dbg"
			print >>f, "Architecture:",packageArchitecture
			writeDependency(f, "Depends", ["${misc:Depends}", binaryName+" (= ${binary:Version})"])
			print >>f, "Description:",sourceName,"debug smbols (auto-debuild)"
			print >>f, " Package containing debug symbols for "+sourceName+", auto-generated by auto-debuild."
			files.append(os.path.join(debDir, "%s-dbg_%s_%s.deb" % (binaryName, version, arch)))
	# install file
	with open('debian/'+binaryName+'.install', 'w') as f:
		for line in config.get('binaryInstallFiles', []):
			if line.startswith('/'): # a file from within the package, not from the source tree
				line = 'debian/'+binaryName+line
			print >>f, line
	# maintainer scripts for alternatives
	if 'alternatives' in config:
		with open('debian/'+binaryName+'.postinst', 'w') as f:
			print >>f, "#!/bin/sh"
			print >>f, "set -e"
			print >>f, 'if [ "$1" = "configure" ]; then'
			for alternative in config['alternatives']:
				print >>f, safeCall('update-alternatives', '--install', alternative['link'], alternative['name'], alternative['target'],
					str(alternative['priority']))
			print >>f, 'fi'
			print >>f, ''
			print >>f, '#DEBHELPER#'
			print >>f, ''
			print >>f, 'exit 0'
		with open('debian/'+binaryName+'.prerm', 'w') as f:
			print >>f, "#!/bin/sh"
			print >>f, "set -e"
			print >>f, 'if [ "$1" = "remove" ]; then'
			for alternative in config['alternatives']:
				print >>f, safeCall('update-alternatives', '--remove', alternative['name'], alternative['target'])
			print >>f, 'fi'
			print >>f, ''
			print >>f, '#DEBHELPER#'
			print >>f, ''
			print >>f, 'exit 0'
	# rules file: build system specific
	with open('debian/rules', 'w') as f:
		# get rule file for build system: may only touch auto_config and auto_clean rules and the dh options
		if buildSystem == 'cmake':
			r = cmakeRules(config)
		elif buildSystem == 'automake':
			r = automakeRules(config)
		else:
			raise Exception("Invalid build system "+buildSystem)
		# global rules
		r.env["DEB_BUILD_OPTIONS"] = 'parallel='+str(parallelJobs)
		if not dbgPackage:
			# disable debug information
			r.env["DEB_CFLAGS_APPEND"] = '-g0'
			r.env["DEB_CXXFLAGS_APPEND"] = '-g0'
		r.dh += ['--parallel']
		if withPython2:
			r.dh += ['--with=python2']
			r.rules['python2'] = ['dh_python2 --no-guessing-versions']
		r.rules['builddeb'] = [safeCall('dh_builddeb', "--destdir="+debDir)] # passing this gobally to dh results in weird problems (like stuff being installed there, and not in the package...)
		r.rules['auto_test'] = []
		# installation rule
		r.rules['auto_install'] = [safeCall('dh_auto_install', '--destdir=debian/'+binaryName)] # install everything into the binary package
		if 'binarySkipFiles' in config:
			r.rules['auto_install'].append(safeCall('cd', 'debian/'+binaryName) + " && " +
				safeCall('rm', *config['binarySkipFiles']))
		# for debug packages
		if dbgPackage:
			r.rules['strip'] = [safeCall('dh_strip', '--dbg-package='+binaryName+"-dbg")] # put debug files in appropriate package
			r.rules['installdocs'] = [safeCall('dh_installdocs', '--link-doc='+binaryName)] # make the doc folder of the dbg package a symlink
		# wait after configuration?
		if config.get('waitAfterConfig', False):
			r.rules['auto_configure'].append("@"+safeCall('read', '-p', 'Configuration done. Hit "Enter" to build the package. ', 'DUMMY_VAR')) # if we run in dash, we need to tell it which variable to use for the result...
		# dump it to a file
		r.write(f)
	mode = os.stat('debian/rules').st_mode
	os.chmod('debian/rules', mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
	# return list of files that will be created
	return files

def buildDebianPackage(config):
	commands = ['dpkg-checkbuilddeps', 'debian/rules clean', 'debian/rules build', 'fakeroot debian/rules binary', 'debian/rules clean']
	command = ['bash', '-c', ' && '.join(commands)] # make it all one command, so we don't have to open and close the chroot too often
	subprocess.check_call(commandInBuildEnv(config, command))
	shutil.rmtree('debian') # it only contains what we just created

###################################################################
# if we are called directly as script
if __name__ == "__main__":
	try:
		import imp
		# get config
		config = imp.load_source('config', 'auto-debuild.conf').__dict__
		os.remove('auto-debuild.confc')
		# generate debian files
		if os.path.exists('debian'):
			if raw_input("A debian folder already exists, to you want to remove it (y/N)? ").lower() != "y":
				sys.exit(1)
			shutil.rmtree('debian')
		files = createDebianFiles(config)
		# check if a file is overwritten
		for file in files:
			if os.path.exists(file):
				if raw_input("Do you want to overwrite %s (y/N)? " % file).lower() != "y":
					sys.exit(1)
		# run compilation
		buildDebianPackage(config)
		# install files
		print "Installing created deb files..."
		subprocess.check_call(['sudo', 'dpkg', '--install'] + files)
	except (subprocess.CalledProcessError, KeyboardInterrupt) as e: # for some exceptions, a stackrace is usually pointless
		print >> sys.stderr
		print >> sys.stderr
		if isinstance(e, KeyboardInterrupt): # str(e) would be the empty string
			print >> sys.stderr, "Interruped by user"
		else:
			print >> sys.stderr, "Error during package creation: %s" % str(e)
		print >> sys.stderr
		sys.exit(1)
